#!/usr/bin/env python3

"""
Helper functions for reading DSMS store.
"""

import json
import requests
import sys
from os import path
from time import sleep
from urllib.parse import urlparse

# if script's parent directory isn't in sys.path, then add it
if ('/'.join(sys.path[0].split("/")[:-1]) not in sys.path):
    sys.path.append('/'.join(sys.path[0].split("/")[:-1]))

from aptools import statsdhelper


class DsmsHelper:
    """
    Helper class to interact with DSMS endpoints.
    """
    # How many times we need to retry and at what delay intervals.
    # DSMS endpoint requires that the delay be exponential.
    RETRY_DELAY_SECS = [0, 1, 2, 4]

    # Cluster to region map. This is available in Shared.ini on the machine.
    # Shared.ini is delivered by data deployment, which makes bootstrap secret dependent
    # on data deployment to be up. Hence we keep the current set of mappings here.
    _cluster_region_map = {"DEFAULT": "uswest2",
                           "apdevccp": "uscentraleuap",
                           "aptest01": "uscentraleuap",
                           "aptest04": "useast2euap",
                           "aptestccp": "useast2euap",
                           "bJ1": "koreacentral",
                           "bn01": "useast2",
                           "bn1Lab": "useast2",
                           "bn2b": "useast2",
                           "bn1": "useast2",
                           "bn2": "useast2",
                           "bn4c": "useast2",
                           "ch01": "usnorth",
                           "chi02": "usnorth",
                           "cO01": "uswest2",
                           "cO1C": "uswest2",
                           "co4": "uswest2",
                           "co4c": "uswest2",
                           "cy2b": "uswestcentral",
                           "cy2test01": "uswestcentral",
                           "cy2test02": "uswestcentral",
                           "cy4": "uswestcentral",
                           "ch1": "usnorth",
                           "ch1b": "usnorth",
                           "ch1d": "usnorth",
                           "co3": "uswest2",
                           "co3b": "uswest2",
                           "dub01": "europenorth",
                           "dub02": "europenorth",
                           "db3": "europenorth",
                           "db4": "europenorth",
                           "dm3": "uscentral",
                           "hk2": "asiaeast",
                           "hkg01": "asiaeast",
                           "mw1": "uswest2",
                           "mwh01": "uswest2",
                           "sg1": "asiasoutheast",
                           "sn2": "ussouth",
                           "co3c": "uswest2",
                           "cy2": "uswestcentral",
                           "db5": "europenorth",
                           "dm2c": "uscentral"}

    def __init__(self, service_name, dsmsdns='', cluster='', cert='', key='', cli=False):
        self.service_name = service_name
        self.cluster = cluster

        self.certificate = cert
        self.key = key

        # We need to handle calls from command line interface differently.
        # For instance, no metric uploads, no retries, no trying default dsms, enable
        # verbose logging etc.,
        self.cli = cli

        # Metric MDM namespace to upload metrics
        if cli is False:
            self.metric_namespace = "APSharedServices"
            self.metrics = statsdhelper.StatsdHelper(statsd_port=9125)
        else:
            self.metrics = None

        if not dsmsdns:
            self.dsmsdns = self.__get_dsms_dns(cluster)
        else:
            self.dsmsdns = dsmsdns

    def __build_dns_string(self, region: str) -> str:
        """
        Build the dsms dns string and return.
        """
        return 'https://{0}-dsms.dsms.core.windows.net'.format(region)

    def __get_dsms_dns(self, cluster: str) -> str:
        """
        Map the ap cluster to a dsms region and return the full dns name.
        If no region mapping exists, the default dns is returned.
        """
        if cluster in self._cluster_region_map:
            region = self._cluster_region_map[cluster.lower()]
        else:
            # Send error metric out. We should alert on this and add a mapping
            # to the missing cluster.
            if self.cli is True:
                raise IOError(
                    "Could not get dsms region for cluster [{0}]. Specify a valid cluster.".format(cluster))
            if self.metrics is not None:
                self.metrics.send_gauge(self.metric_namespace,
                                        "DSMSClusterRegionMappingMissing",
                                        1)
            return self.__get_default_dsms_dns()

        return self.__build_dns_string(region)

    def __get_default_dsms_dns(self) -> str:
        """
        In instances where we can't find a dsms region mapping for an AP cluster,
        try to use a default dsms region.
        """
        region = self._cluster_region_map["DEFAULT"]
        return self.__build_dns_string(region)

    def _get_dsms_url(self) -> str:
        """
        Returns the full URL for the REST API call to be made to DSMS
        Format: https://<DsmsDns>/pfagent/{ServiceName}?agentless 
        e.g. https://uswest2-dsms.core.windows.net/pfagent/aplinux/services/pfshared/co4?agentless 
        """
        url_path = '{}/pfagent/{}?agentless'.format(self.dsmsdns, self.service_name)
        return url_path

    def __get_default_url(self) -> str:
        """
        Returns the full URL for the REST API call to be made to default DSMS region.
        Format: https://<DsmsDns>/pfagent/{ServiceName}?agentless 
        e.g. https://uswest2-dsms.core.windows.net/pfagent/aplinux/services/pfshared/co4?agentless 
        """
        default_dsms_dns = self.__get_default_dsms_dns()
        url_path = '{}/pfagent/{}?agentless'.format(default_dsms_dns, self.service_name)

        return url_path

    def _call_dsms_rest_api(self, api: str) -> requests.Response:
        """
        Calls the DSMS rest API and return the response.
        """
        # Ensure we can read the client certs
        if path.isfile(self.certificate) is False or path.isfile(self.key) is False:
            raise IOError(
                "Cert/Key file not found cert={0} key={1}".format(self.certificate, self.key))

        client_cert = (self.certificate, self.key)
        url = self._get_dsms_url()

        if api == "GetAllSecrets":
            return self._dsms_api_get_all_secrets(url, client_cert)
        else:
            raise ValueError("{0} is not a valid DSMS api".format(api))

    def _dsms_api_get_all_secrets(self, url: str, cert) -> requests.Response:
        """
        Send a GetAllSecrets command to dsms endpoint.
        """
        # HTTP Header - To close the connection immediately after completion of the response.
        # DSMS gets a lot of responses, don't keep the conn open.
        header = {'Connection': 'close'}

        # Try connecting to DSMS more than once. Use exponential delay seconds. Required
        # by dsms team. If couldn't read from assigned dsms region, try the default region
        # once.
        assigned_region_exception = None  # Var to save exception from assigned region, if any.
        for attempt, delay in enumerate(self.RETRY_DELAY_SECS):
            # Delay our attempt by specified seconds. 0 is the first delay.
            sleep(delay)

            # If this is the last attempt
            last_attempt = (attempt == len(self.RETRY_DELAY_SECS) - 1)
            if last_attempt:
                # Replace url to try the default endpoint.
                url = self.__get_default_url()
                print("WARNING: DSMS - trying default endpoint: {0}".format(url), file=sys.stderr)

                # Send metric up indicating the issue connecting to assigned dsms region.
                if self.metrics is not None:
                    self.metrics.send_gauge(self.metric_namespace,
                                            "DSMSConnectivityError",
                                            1,
                                            {'DsmsDns': self.dsmsdns})
            try:
                # Send the request. GetAllSecrets = GET /servicepath/
                response = requests.get(url, cert=cert, timeout=5, headers=header)
                return response
            except Exception as e:
                # Catch exception to attempt call again.
                if self.cli is True:
                    # If this is from a cli, we don't do multiple attempts or try backup
                    # dsms endpoint.
                    raise IOError("Failed to connect to DSMS: {0}".format(e))

                print(
                    "WARNING: DSMS - Connection attempt {0} failed: {1}\nWill retry after {2} secs - {3}".format(
                        attempt +
                        1,
                        url,
                        delay,
                        e),
                    file=sys.stderr)

                # If this was the last attempt, propagate the previous error out. Current exception
                # pertains to the default endpoint. We want to propagate error reaching assigned
                # one.
                if last_attempt:
                    raise IOError(
                        "Failed to connect to DSMS: {0}".format(assigned_region_exception))
                else:
                    # Store the last exception to propagate if even last attempt using default
                    # dsms fails.
                    assigned_region_exception = e

    def get_all_secrets(self) -> str:
        """
        Returns the value of the specified dSMS secret or an empty string if the secret doesn't exist.
        Connect and issue a GET on the endpoint. There is only one API GetAllSecrets available for
        JSON APPKI endpoint.
        :return: json string
        """
        # Get response from dsms
        try:
            response = self._call_dsms_rest_api('GetAllSecrets')

            # Raise an exception if the HTTP response indicates an issue.
            # 200 is the valid response, otherwise we error out.
            response.raise_for_status()

            resp_text = response.text

            response.close()

            # Raise exception if a valid json wasn't sent back.
            json.loads(resp_text)

        except Exception as e:
            if self.metrics is not None:
                self.metrics.send_gauge(self.metric_namespace,
                                        "DSMSReadError",
                                        1,
                                        {'DsmsDns': self.dsmsdns})
            raise IOError("Error: Reading from DSMS: {0}".format(e))

        return resp_text


class DsmsResponse(object):
    """
    Store and process a dsms response json object.
    """

    # Response JSON contains different sections for each type of secret.
    AdHocSecretObjName = "AdHocSecretInfos"
    CertObjName = "CertInfos"
    AccountSasObjName = "StorageAccountInfos"
    SourceLocationObjName = "SourceLocation"
    AdhocSecretFieldName = "Secret"
    CertPfxFieldName = "CertificatePfx"
    AdhocCurrentVersionName = "CurrentAdHocSecret"
    CertCurrentVersionName = "CurrentCertificateSecret"

    # Each secret is stored in different path based on its type in dSMS.
    secret_json_object_map = {
        "adhocsecrets": AdHocSecretObjName,
        "certificates": CertObjName,
        "accountsas": AccountSasObjName
    }

    def __init__(self, namespace: str, response_json: str):
        if not namespace:
            raise ValueError("Empty namespace")

        if not response_json:
            raise ValueError("Empty response json")

        self.namespace = namespace

        # This will raise an exception if valid JSON wasn't passed in. Caller should catch it.
        self.json_dict = json.loads(response_json)

    def get_secret(self, name: str) -> str:
        """
        Get a secret from a loaded response.
        """
        if not name:
            raise ValueError("Get secret: Empty secret name")

        # Each secret type is stored in it's own section in the json object
        # Name will be of format <secret-type>/path/to/secret
        secret_type = self._get_secret_type(name)

        if secret_type == self.AdHocSecretObjName:
            return self.__get_adhoc_secret(name)
        elif secret_type == self.CertObjName:
            return self.__get_certificate(name)
        else:
            raise ValueError("Secret type for '{0}' could not be mapped".format(name))

    def __get_full_secret_path(self, name: str) -> str:
        """
        Each secret object has a SourceLocation that contains the path to the secret. This
        item is the only unique string we can use to search our target string.
        e.g. "SourceLocation": "https://uscentraleuap-dsms.dsms.core.windows.net/dsms/aplinux/adhocsecrets/test/acr/password1",
        """
        return "/dsms/{0}/{1}".format(self.namespace, name)

    def __get_adhoc_secret(self, name: str) -> str:
        """
        Get an secret of type Ad Hoc. Ad hoc secrets start with path name /adhocsecrets/...
        :return: Secret in string format
        """
        if self.AdHocSecretObjName not in self.json_dict:
            raise IOError(
                "Could not find {0} section in loaded response".format(
                    self.AdhocSecretFieldName))

        # Pick up the ad hoc array.
        adhoc_json_array = self.json_dict[self.AdHocSecretObjName]

        # Go through the array searching for our secret using full dsms path
        secret_path = self.__get_full_secret_path(name)
        for secret_obj in adhoc_json_array:
            # e.g.
            # "https://uscentraleuap-dsms.dsms.core.windows.net/dsms/aplinux/adhocsecrets/test/acr/password1",
            source_location_url = secret_obj[self.SourceLocationObjName]

            # Parse and get the url path. The dns entry will change for each cluster/region combo.
            # So we need to look for the url path matching the secret we are searching.
            parsed_url = urlparse(source_location_url)
            if secret_path == parsed_url.path:
                # "CurrentAdHocSecret": {
                #    "Version": 2,
                #    "Secret": "sriramshauthtest-updated\r\nKBGaf=tG5sSvFneVt5E+ERjbsivNonYW\r\n"}
                current_version = secret_obj[self.AdhocCurrentVersionName]
                return current_version[self.AdhocSecretFieldName]

        raise IOError("Could not find adhoc secret '{0}' in response.".format(name))

    def __get_certificate(self, name: str) -> str:
        """
        Get a certificate. Certificates start with path name /certificates/...
        :return: Cert in string format
        """
        if self.CertObjName not in self.json_dict:
            raise IOError(
                "Could not find {0} section in loaded response".format(
                    self.AdhocSecretFieldName))

        # Pick up the cert array.
        cert_json_array = self.json_dict[self.CertObjName]

        # Go through the array searching for our secret using full dsms path
        secret_path = self.__get_full_secret_path(name)
        for secret_obj in cert_json_array:
            # e.g.
            # "https://uscentraleuap-dsms.dsms.core.windows.net/dsms/aplinux/adhocsecrets/test/acr/password1",
            source_location_url = secret_obj[self.SourceLocationObjName]

            # Parse and get the url path. The dns entry will change for each cluster/region combo.
            # So we need to look for the url path matching the secret we are searching.
            parsed_url = urlparse(source_location_url)
            if secret_path == parsed_url.path:
                current_version = secret_obj[self.CertCurrentVersionName]
                return current_version[self.CertPfxFieldName]

        raise IOError("Could not find certificate '{0}' in response.".format(name))

    def _get_secret_type(self, name: str) -> str:
        # Each secret type is stored in it's own section in the json object
        # Name will be of format <secret-type>/path/to/secret
        # e.g. adhocsecrets/path/to/secret
        # Strip leading '/' char
        type_str = name.split('/')[0]
        if type_str in self.secret_json_object_map:
            return self.secret_json_object_map[type_str]
        else:
            return None
